from dagster_aws.components.athena import (
    AthenaClientResourceComponent as AthenaClientResourceComponent,
)
from dagster_aws.components.credentials import (
    AthenaCredentialsComponent as AthenaCredentialsComponent,
    Boto3CredentialsComponent as Boto3CredentialsComponent,
    S3CredentialsComponent as S3CredentialsComponent,
)
from dagster_aws.components.ecr import ECRPublicResourceComponent as ECRPublicResourceComponent
from dagster_aws.components.rds import RDSResourceComponent as RDSResourceComponent
from dagster_aws.components.s3 import (
    S3FileManagerResourceComponent as S3FileManagerResourceComponent,
    S3ResourceComponent as S3ResourceComponent,
)
from dagster_aws.components.secretsmanager import (
    SecretsManagerResourceComponent as SecretsManagerResourceComponent,
    SecretsManagerSecretsResourceComponent as SecretsManagerSecretsResourceComponent,
)
from dagster_aws.components.ssm import (
    ParameterStoreResourceComponent as ParameterStoreResourceComponent,
    SSMResourceComponent as SSMResourceComponent,
)



import json
from collections.abc import Mapping
from typing import Any

import boto3
from dagster import PipesClient
from dagster._annotations import public
from dagster._core.definitions.resource_annotation import TreatAsResourceParam
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.execution.context.compute import OpExecutionContext
from dagster._core.pipes.client import (
    PipesClientCompletedInvocation,
    PipesContextInjector,
    PipesMessageReader,
)
from dagster._core.pipes.utils import open_pipes_session

from dagster_aws.pipes.context_injectors import PipesLambdaEventContextInjector
from dagster_aws.pipes.message_readers import PipesLambdaLogsMessageReader


class PipesLambdaClient(PipesClient, TreatAsResourceParam):
    """A pipes client for invoking AWS lambda.

    By default context is injected via the lambda input event and messages are parsed out of the
    4k tail of logs.

    Args:
        client (boto3.client): The boto lambda client used to call invoke.
        context_injector (Optional[PipesContextInjector]): A context injector to use to inject
            context into the lambda function. Defaults to :py:class:`PipesLambdaEventContextInjector`.
        message_reader (Optional[PipesMessageReader]): A message reader to use to read messages
            from the lambda function. Defaults to :py:class:`PipesLambdaLogsMessageReader`.
    """

    def __init__(
        self,
        client: "boto3.client | None" = None,  # pyright: ignore (reportGeneralTypeIssues)
        context_injector: PipesContextInjector | None = None,
        message_reader: PipesMessageReader | None = None,
    ):
        self._client = client or boto3.client("lambda")
        self._message_reader = message_reader or PipesLambdaLogsMessageReader()
        self._context_injector = context_injector or PipesLambdaEventContextInjector()

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    @public
    def run(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        *,
        function_name: str,
        event: Mapping[str, Any],
        context: OpExecutionContext | AssetExecutionContext,
    ) -> PipesClientCompletedInvocation:
        """Synchronously invoke a lambda function, enriched with the pipes protocol.

        Args:
            function_name (str): The name of the function to use.
            event (Mapping[str, Any]): A JSON serializable object to pass as input to the lambda.
            context (Union[OpExecutionContext, AssetExecutionContext]): The context of the currently executing Dagster op or asset.

        Returns:
            PipesClientCompletedInvocation: Wrapper containing results reported by the external
            process.
        """
        with open_pipes_session(
            context=context,
            message_reader=self._message_reader,
            context_injector=self._context_injector,
        ) as session:
            other_kwargs = {}
            if isinstance(self._message_reader, PipesLambdaLogsMessageReader):
                other_kwargs["LogType"] = "Tail"

            if isinstance(self._context_injector, PipesLambdaEventContextInjector):
                payload_data = {
                    **event,
                    **session.get_bootstrap_env_vars(),
                }
            else:
                payload_data = event

            response = self._client.invoke(
                FunctionName=function_name,
                InvocationType="RequestResponse",
                Payload=json.dumps(payload_data),
                **other_kwargs,
            )
            if isinstance(self._message_reader, PipesLambdaLogsMessageReader):
                self._message_reader.consume_lambda_logs(response)

            if "FunctionError" in response:
                err_payload = json.loads(response["Payload"].read().decode("utf-8"))

                raise Exception(
                    f"Lambda Function Error ({response['FunctionError']}):\n{json.dumps(err_payload, indent=2)}"
                )

        # should probably have a way to return the lambda result payload
        return PipesClientCompletedInvocation(session)



from functools import cached_property

import dagster as dg
from dagster._annotations import preview, public
from pydantic import Field

from dagster_aws.athena.resources import AthenaClientResource
from dagster_aws.components.credentials import AthenaCredentialsComponent


@public
@preview
class AthenaClientResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides an AthenaClientResource for executing queries against Amazon Athena."""

    credentials: AthenaCredentialsComponent = Field(
        description="Credentials for connecting to Athena."
    )

    resource_key: str | None = Field(
        default=None,
        description="The key under which the Athena resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> AthenaClientResource:
        """Resolves credentials and returns a configured Athena resource."""
        return AthenaClientResource(**self.credentials.render_as_dict())

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        """Binds the Athena resource to the specified resource key in the definitions."""
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})



from functools import cached_property

import dagster as dg
from dagster._annotations import preview, public
from pydantic import Field

from dagster_aws.components.credentials import S3CredentialsComponent
from dagster_aws.s3.resources import S3FileManagerResource, S3Resource


@public
@preview
class S3ResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides an S3Resource for interacting with AWS S3."""

    credentials: S3CredentialsComponent = Field(description="Credentials for connecting to S3.")

    resource_key: str | None = Field(
        default=None,
        description="The key under which the S3 resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> S3Resource:
        """Resolves credentials and returns a configured S3 resource."""
        return S3Resource(**self.credentials.render_as_dict())

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})


@public
@preview
class S3FileManagerResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides an S3FileManagerResource for file management operations on AWS S3."""

    s3_bucket: str = Field(description="S3 bucket to use for the file manager.")
    s3_prefix: str = Field(
        default="dagster", description="Prefix to use for the S3 bucket for this file manager."
    )

    credentials: S3CredentialsComponent = Field(description="Credentials for connecting to S3.")

    resource_key: str | None = Field(
        default=None,
        description="The key under which the S3FileManager resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> S3FileManagerResource:
        """Resolves credentials and returns a configured S3FileManagerResource."""
        return S3FileManagerResource(
            s3_bucket=self.s3_bucket, s3_prefix=self.s3_prefix, **self.credentials.render_as_dict()
        )

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})


S3ResourceComponent.model_rebuild()

S3FileManagerResourceComponent.model_rebuild()



from datetime import datetime
from typing import Any, cast

import boto3
import dagster._check as check
from dagster._annotations import deprecated

from dagster_aws._stubs import ObjectTypeDef


class ClientException(Exception):
    pass


def get_objects(
    bucket: str,
    prefix: str = "",
    since_key: str | None = None,
    since_last_modified: datetime | None = None,
    client=None,
) -> list[ObjectTypeDef]:
    """Retrieves a list of object keys in S3 for a given `bucket`, `prefix`, and filter option.

    Args:
        bucket (str): s3 bucket
        prefix (str): s3 object prefix
        since_key (Optional[str]): retrieve objects modified after the last modified timestamp of this key
        since_last_modified (Optional[datetime]): retrieve objects after this timestamp (non-inclusive)
        client (Optional[boto3.Client]): s3 client

    Returns:
        List of object keys in S3.

    """
    check.str_param(bucket, "bucket")
    check.str_param(prefix, "prefix")
    check.opt_str_param(since_key, "since_key")
    check.opt_inst_param(since_last_modified, "since_last_modified", datetime)

    if not client:
        client = boto3.client("s3")

    if not client:
        raise ClientException("Failed to initialize s3 client")

    paginator = client.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix, Delimiter="")

    objects: list[ObjectTypeDef] = []
    for page in page_iterator:
        contents = page.get("Contents", [])
        objects.extend([cast("ObjectTypeDef", obj) for obj in contents])

    if since_key and not any(obj.get("Key") == since_key for obj in objects):
        raise Exception("Provided `since_key` is not present in list of objects")

    sorted_objects = [obj for obj in sorted(objects, key=lambda x: x.get("LastModified"))]

    if since_key:
        for idx, obj in enumerate(sorted_objects):
            if obj.get("Key") == since_key:
                return sorted_objects[idx + 1 :]

    if since_last_modified:
        for idx, obj in enumerate(sorted_objects):
            if obj.get("LastModified") > since_last_modified:
                return sorted_objects[idx:]
        # If no new files are found, return an empty list.
        return []

    return sorted_objects


@deprecated(breaking_version="2.0", additional_warn_text="Use get_objects instead.")
def get_s3_keys(
    bucket: str,
    prefix: str = "",
    since_key: str | None = None,
    s3_session: Any | None = None,
) -> list[str]:
    """Retrieves a list of object keys in S3 for a given `bucket`, `prefix`, and filter option.

    Note: when using the `since_key` it is possible to miss records if that key has been modified,
    as sorting is done by the `LastModified` property of the S3 object. For more information, see
    the following GitHub issue:

        https://github.com/dagster-io/dagster/issues/22892

    Args:
        bucket (str): s3 bucket
        prefix (str): s3 object prefix
        since_key (Optional[str]): retrieve objects after the modified date of this key
        since_last_modified (Optional[str]): retrieve objects after this timestamp
        s3_session (Optional[boto3.Client]): s3 client

    Returns:
        List of object keys in S3.

    """
    objects = get_objects(bucket=bucket, prefix=prefix, since_key=since_key, client=s3_session)
    return [obj.get("Key") for obj in objects]



import hashlib
import re
from collections.abc import Mapping
from typing import Any

from dagster._core.remote_origin import RemoteJobOrigin

from dagster_aws.ecs.tasks import DagsterEcsTaskDefinitionConfig


def sanitize_family(family):
    # Trim the location name and remove special characters
    return re.sub(r"[^\w^\-]", "", family)[:255]


def sanitize_tag(tag):
    # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/ecs-using-tags.html
    return re.sub("[^A-Za-z0-9_]", "-", tag)[:255].strip("-")


def _get_family_hash(name):
    m = hashlib.sha1()
    m.update(name.encode("utf-8"))
    name_hash = m.hexdigest()[:8]
    return f"{name[:55]}_{name_hash}"


class RetryableEcsException(Exception): ...


def run_ecs_task(ecs, run_task_kwargs) -> Mapping[str, Any]:
    try:
        response = ecs.run_task(**run_task_kwargs)
    except ecs.exceptions.InvalidParameterException as e:
        # ECS sometimes raises transient throttling errors in the underlying EC2 services
        # as InvalidParameterExceptions
        if "Throttling" in e.response.get("Error", {}).get("Message", ""):
            raise RetryableEcsException(str(e)) from e
        raise

    tasks = response["tasks"]

    if not tasks:
        failures = response["failures"]
        failure_messages = []
        for failure in failures:
            arn = failure.get("arn")
            reason = failure.get("reason")
            detail = failure.get("detail")

            failure_message = (
                "Task"
                + (f" {arn}" if arn else "")
                + " failed."
                + (f" Failure reason: {reason}" if reason else "")
                + (f" Failure details: {detail}" if detail else "")
            )
            failure_messages.append(failure_message)

        failure_message = "\n".join(failure_messages) if failure_messages else "Task failed."

        if "Capacity is unavailable at this time" in failure_message:
            raise RetryableEcsException(failure_message)

        raise Exception(failure_message)
    return tasks[0]


def get_task_definition_family(
    prefix: str,
    job_origin: RemoteJobOrigin,
) -> str:
    job_name = job_origin.job_name
    repo_name = job_origin.repository_origin.repository_name
    location_name = job_origin.repository_origin.code_location_origin.location_name

    assert len(prefix) < 32

    # Truncate the location name if it's too long (but add a unique suffix at the end so that no matter what it's unique)
    # Relies on the fact that org name and deployment name are always <= 64 characters long to
    # stay well underneath the 255 character limit imposed by ECS

    final_family = f"{prefix}_{_get_family_hash(location_name)}_{_get_family_hash(repo_name)}_{_get_family_hash(job_name)}"

    assert len(final_family) <= 255

    return sanitize_family(final_family)


def task_definitions_match(
    desired_task_definition_config: DagsterEcsTaskDefinitionConfig,
    existing_task_definition: Mapping[str, Any],
    container_name: str,
) -> bool:
    if not any(
        [
            container["name"] == container_name
            for container in existing_task_definition["containerDefinitions"]
        ]
    ):
        return False

    existing_task_definition_config = DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
        existing_task_definition, container_name
    )

    return existing_task_definition_config.matches_other_task_definition_config(
        desired_task_definition_config
    )


def get_task_logs(ecs, logs_client, cluster, task_arn, container_name, limit=10):
    task = ecs.describe_tasks(cluster=cluster, tasks=[task_arn]).get("tasks")[0]

    task_definition_arn = task.get("taskDefinitionArn")
    task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn).get(
        "taskDefinition"
    )

    matching_container_definitions = [
        container_definition
        for container_definition in task_definition.get("containerDefinitions", [])
        if container_definition["name"] == container_name
    ]
    if not matching_container_definitions:
        raise Exception(f"Could not find container with name {container_name}")

    container_definition = matching_container_definitions[0]

    log_options = container_definition.get("logConfiguration", {}).get("options", {})
    log_group = log_options.get("awslogs-group")
    log_stream_prefix = log_options.get("awslogs-stream-prefix")

    if not log_group or not log_stream_prefix:
        return []

    container_name = container_definition.get("name")
    task_id = task_arn.split("/")[-1]

    log_stream = f"{log_stream_prefix}/{container_name}/{task_id}"

    events = logs_client.get_log_events(
        logGroupName=log_group, logStreamName=log_stream, limit=limit
    ).get("events")

    return [event.get("message") for event in events]


def is_transient_task_stopped_reason(stopped_reason: str) -> bool:
    if "Timeout waiting for network interface provisioning to complete" in stopped_reason:
        return True

    if "Timeout waiting for EphemeralStorage provisioning to complete" in stopped_reason:
        return True

    if "CannotPullContainerError" in stopped_reason and "i/o timeout" in stopped_reason:
        return True

    if "CannotPullContainerError" in stopped_reason and (
        "invalid argument" in stopped_reason or "EOF" in stopped_reason
    ):
        return True

    if (
        "Unexpected EC2 error while attempting to Create Network Interface" in stopped_reason
        and "AuthFailure" in stopped_reason
    ):
        return True

    if "The Service Discovery instance could not be registered" in stopped_reason:
        return True

    return False



from functools import cached_property

import dagster as dg
from dagster._annotations import preview, public
from pydantic import Field

from dagster_aws.components.credentials import Boto3CredentialsComponent
from dagster_aws.rds.resources import RDSResource


@public
@preview
class RDSResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides an RDSResource for interacting with AWS RDS service."""

    credentials: Boto3CredentialsComponent = Field(
        description="AWS credentials - inline configuration."
    )
    resource_key: str | None = Field(
        default=None,
        description="The key under which the RDS resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> RDSResource:
        """Resolves credentials and returns a configured RDS resource."""
        return RDSResource(**self.credentials.render_as_dict())

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})



from dagster_shared.libraries import DagsterLibraryRegistry

from dagster_aws.components import (
    AthenaClientResourceComponent as AthenaClientResourceComponent,
    AthenaCredentialsComponent as AthenaCredentialsComponent,
    Boto3CredentialsComponent as Boto3CredentialsComponent,
    ECRPublicResourceComponent as ECRPublicResourceComponent,
    ParameterStoreResourceComponent as ParameterStoreResourceComponent,
    RDSResourceComponent as RDSResourceComponent,
    S3CredentialsComponent as S3CredentialsComponent,
    S3FileManagerResourceComponent as S3FileManagerResourceComponent,
    S3ResourceComponent as S3ResourceComponent,
    SecretsManagerResourceComponent as SecretsManagerResourceComponent,
    SecretsManagerSecretsResourceComponent as SecretsManagerSecretsResourceComponent,
    SSMResourceComponent as SSMResourceComponent,
)
from dagster_aws.version import __version__

DagsterLibraryRegistry.register("dagster-aws", __version__)



from functools import cached_property

import dagster as dg
from dagster._annotations import preview, public
from pydantic import Field

from dagster_aws.components.credentials import Boto3CredentialsComponent
from dagster_aws.secretsmanager.resources import (
    SecretsManagerResource,
    SecretsManagerSecretsResource,
)


@public
@preview
class SecretsManagerResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides a SecretsManagerResource for interacting with AWS Secrets Manager."""

    credentials: Boto3CredentialsComponent = Field(
        description="AWS credentials - inline configuration."
    )
    resource_key: str | None = Field(
        default=None,
        description="The key under which the SecretsManager resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> SecretsManagerResource:
        """Resolves credentials and returns a configured SecretsManager resource."""
        return SecretsManagerResource(**self.credentials.render_as_dict())

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})


@public
@preview
class SecretsManagerSecretsResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides a SecretsManagerSecretsResource for fetching secrets from AWS Secrets Manager."""

    credentials: Boto3CredentialsComponent = Field(
        description="AWS credentials - inline configuration."
    )
    secrets: list[str] = Field(
        default=[], description="An array of AWS Secrets Manager secrets ARNs to fetch."
    )
    secrets_tag: str | None = Field(
        default=None,
        description="AWS Secrets Manager secrets with this tag will be fetched and made available.",
    )
    resource_key: str | None = Field(
        default=None,
        description="The key under which the SecretsManagerSecrets resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> SecretsManagerSecretsResource:
        creds_data = self.credentials.render_as_dict()
        return SecretsManagerSecretsResource(
            **creds_data,
            secrets=self.secrets,
            secrets_tag=self.secrets_tag,
        )

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})
